/*
 * @(#)JavaCGIBridge.java 2.00 07/18/98
 *
 * Copyright Info: This class was written by Gunther Birznieks
 * gunther@clark.net having been inspired by countless other
 * software developers.
 *
 * Version 1.00: 08/10/97
 * Version 2.00: 07/19/98
 *
 * The second version of this software was written for
 * independent study under the guidance of Professor 
 * Marty Hall as part of the Johns Hopkins University 
 * Computer Science Master's Degree program.
 *
 * Feel free to copy, cite, reference, sample, borrow, resell
 * or plagiarize the contents.  However, if you don't mind,
 * please let me know where it goes so that I can at least
 * watch and take part in the development of the memes. Information
 * wants to be free; support public domain freeware.  Donations are
 * appreciated and will be spent on further upgrades and other public
 * domain programs.
 *
 */

package com.extropia.net;

import java.util.Hashtable;
import java.util.Enumeration;
import java.util.Vector;

import java.net.URL;
import java.net.URLConnection;
import java.net.URLEncoder;
import java.net.MalformedURLException;

import java.io.DataOutputStream;
import java.io.DataInputStream;
import java.io.IOException;

/**
 * This is a class that POSTS and GETS data from URLs using various
 * helper methods.  This class also provides the capability of
 * timing out if the connection takes too long to transfer data
 * by using threads to monitor whether the data is taking too long
 * to transfer (implements Runnable).
 * <P>
 * Helper methods allow you to set up form variables, get setup
 * information over a URL, get raw or pre-parsed HTML data, and more.
 * <P>
 * The getParsedData method relies on instance variables which
 * tell the parser where the data begins and ends (top and bottom
 * separators respectively).   The field and row separators inform
 * the parser where records and fields end.
 * <P><BLOCKQUOTE>
 * The default top separator is &LT!--start of data-->
 * <P>
 * The default bottom separator is &LT!--end of data-->
 * <P>
 * The default field separator is ~|~ (tilde + pipe + tilde).
 * <P>
 * The default row separator is ~|~\n (tilde + pipe + tilde + newline).
 * </BLOCKQUOTE>
 *
 * @version     2.00, 18 Jul 1998
 * @author      Gunther Birznieks
 * @author      Marty Hall
 *
 */
public class JavaCGIBridge implements Runnable extends Observable {

    /**
     * The field separator.  When a CGI script or HTML file
     * returns data, then the getParsedData() method will
     * know how to separate fields by looking at this variable.
     *
     * The default value is "~|~" (tilde + pipe + tilde). This is
     * considered sufficiently unlikely to appear inside of actual
     * field data that it is a pretty good field separator to parse
     * on. If the user of this class needs a different separator, it
     * may be overridden by calling the appropriate method.
     */
    private String _fieldSeparator = "~|~";

    /**
     * The row separator.  When a CGI script or HTML file
     * returns data, then the getParsedData() method will
     * know how to separate returned rows by looking at this variable.
     *
     * The default value is "~|~\n" (tilde + pipe + tilde + newline). 
     * This is considered sufficiently unlikely to appear inside of
     * actual row data that it is a pretty good row separator to parse
     * on. If the user of this class needs a different separator, it
     * may be overridden by calling the appropriate method.
     */
    private String _recordSeparator = "~|~\n";

    /**
     * The top of data separator.  When a CGI script or HTML file
     * returns data, then the getParsedData() method will
     * determine when to start parsing data by encountering this
     * separator.
     *
     * The default value is "<!--start of data-->" implicitly followed
     * by a newline.  That is, a newline is generally expected to follow
     * this separator.
     */
    private String _startDataSeparator = "<!--start of data-->\n";

    /**
     * The bottom of data separator.  When a CGI script or HTML file
     * returns data, then the getParsedData() method will
     * determine when to stop parsing data by encountering this
     * separator.
     *
     * The default value is "<!--end of data-->" implicitly followed
     * by a newline.  That is, a newline is generally expected to follow
     * this separator.
     */
    private String _endDataSeparator = "<!--end of data-->\n";

    /**
     * The URL That data is retrieved from.  This is an instance
     * variable rather than a parameter because the data retrieval
     * is done from a launched thread which gets no parameters.
    */
    private URL _threadURL = null;

    /**
     * The HTML form data for the URL that data is retrieved from.  
     * This is an instance variable rather than a parameter because
     * the data retrieval is done from a launched thread which gets
     * no parameters.
    */
    private Hashtable _threadFormVar = null;

    /**
     * This is a flag indicating whether the URL data was
     * retrieved inside the thread or not.
    */
    private boolean _threadCompleted = false;

    /**
     * This is the returned URL data. It is an instance variable because
     * the run() method of the thread cannot explicitly return data.
     */
    private String _threadCGIData = null;

    /**
     * This is the default CGI Timeout in milliseconds.  For example,
     * a value of 10000 will tell the class to throw a
     * JavaCGIBridgeTimeOutException if the data is not retrieved within
     * 10 seconds of having initiated a data transfer.
     */
    private static int _defaultThreadJavaCGIBridgeTimeOut = 10000;

    /**
     * This is the actual CGI Timeout value in milliseconds for the
     * instantiated object. Notice that the default value is a static
     * class variable that applies across all instances of this class.
     * This variable, on the other hand, is the actual value the object
     * uses.
     */
    private int threadJavaCGIBridgeTimeOut = _defaultThreadJavaCGIBridgeTimeOut;

    /** 
     * This is a flag that makes it so that the run() method is
     * at least semi-private for Threading.  This flag has to be
     * set true in order for another object to execute the run() method.
     */
    private boolean _threadCanRunMethodExecute = false;
 
    /** 
     * _parseDataAsVector is a flag letting the HTTP parser know to
     * parse data on the fly as it is coming in.
     *
     */
    private boolean _parseDataAsVector = false;
    private Vector _vectorParsedData = null;

    private int _numberOfHtmlLinesToNotifyObserver = 0;
    private int _numberOfRecordsToNotifyObserver = 0;
    private boolean _notifyObserversOfEndOfData = false;
    private boolean _notifyObserversOfEndOfHtml = false;
    private boolean _notifyObserversOfEndOfVector = false;


    public final int SETUP = 1;
    public final int PROCESSING = 2;
    public final int IDLE = 3;
    private int _javaCGIBridgeStatus = SETUP;

    /**
     * Constructs the object with new separator values
     */
    public JavaCGIBridge(String field, String row, String top, String bottom) {
        _fieldSeparator = new String(field);
        _recordSeparator = new String(row);
        _startDataSeparator = new String(top);
        _endDataSeparator = new String(bottom);
    }

    /**
     * Constructs the object with no initialization. This is the default
     * (empty) constructor.
     */
    public JavaCGIBridge() {
    } // Empty constructor

    /**
     * Adds a form variable, value pair to the passed Hashtable.
     *
     * @param ht the Hashtable that contains the form variable/value pairs
     * @param formKey the String that contains the form variable to add
     * @param formValue the String that contains the form value to add
     */
    public void addFormValue(Hashtable ht, String formKey, String formValue) {
        Vector vValues = null;

        if (formValue != null && formValue.length() > 0) {
            if (ht.containsKey(formKey)) {
                vValues = (Vector)ht.get(formKey);
            } else {
                vValues = new Vector();
            }
            vValues.addElement(formValue);
            ht.put(formKey, vValues);
        }
    } // End of addFormValue

    /** 
     * Takes the parsed data returned from the getParsedData method
     * and changes it to a Hashtable of key, value pairs where the first
     * Vector entry of each Vector record is the key and the rest of the
     * second Vector entry for each record is the value of the Hashtable.
     *
     * @param vectorOfVectors the Vector of Vectors for the parsed data
     * @return Hashtable containing converted variable/value pairs
     * @see #getParsedData
     */ 
    public Hashtable getKeyValuePairs(Vector vectorOfVectors) {
        Hashtable h = new Hashtable();
        Vector v = null;

        for (Enumeration e = vectorOfVectors.elements(); e.hasMoreElements();) {
            v = (Vector)e.nextElement();
            h.put((String)v.elementAt(0), (String)v.elementAt(1));
        }
        return h;
    } // End of getKeyValuePairs

    /**
     * Returns parsed data in the form of a Vector of Vectors containing
     * the returned fields inside of a Vector of returned rows.
     *
     * @param u URL to get parsed data from.
     * @return Vector (records) of vectors (fields) of parsed data
     * @exception JavaCGIBridgeTimeOutException If the retrieval times out
     * @see #getRawCGIData
     */
    public Vector getParsedData(URL u) throws JavaCGIBridgeTimeOutException {
        return (getParsedData(u, null));
    }

    /**
     * Returns parsed data in the form of a Vector of Vectors containing
     * the returned fields inside of a Vector of returned rows. This form
     * POSTs the HTML Form variable data to the URL.
     *
     * @param u URL to get parsed data from.
     * @param ht Hashtable contains form variables to POST
     * @return Vector (records) of vectors (fields) of parsed data
     * @exception JavaCGIBridgeTimeOutException If the retrieval times out
     * @see #getRawCGIData
     */
    public Vector getParsedData(URL u, Hashtable ht) 
      throws JavaCGIBridgeTimeOutException {

        // getParsedData calls getRawCGIData to obtain
        // the actual HTML text

        // Notice that we turn on the flag to parse
        // vector data inside the getRawCGIData block
        // 
        // We need to do this in order to provide a 
        // mechanism for returning partial results to
        // an observer.
        //
        _parseDataAsVector = true;
        String data = getRawCGIData(u,ht);
        _parseDataAsVector = false;

        return _vectorParsedData;
    }

    /**
     * Returns raw HTML data as a String from the passed URL.
     *
     * @param u URL to get raw HTML from.
     * @return String containing plain HTML text
     * @exception JavaCGIBridgeTimeOutException If the retrieval times out
     * @see #getParsedData
     */
    public String getRawCGIData(URL u) throws JavaCGIBridgeTimeOutException {
        return (getRawCGIData(u,null));
    }

    /**
     * Returns raw HTML data as a String from the passed URL and list
     * of Form variable/value pairs stored in a Hashtable.  This form
     * POSTs the HTML Form variable data to the URL.
     *
     * @param u URL to get raw HTML from.
     * @param ht Hashtable contains form variables to POST
     * @return String containing plain HTML text
     * @exception JavaCGIBridgeTimeOutException If the retrieval times out
     * @see #getParsedData
     */
    public String getRawCGIData(URL u, Hashtable ht)
      throws JavaCGIBridgeTimeOutException 
    {
        // We set up the information for passing to the thread
        // ahead of time.
        _threadURL = u;
        _threadFormVar = ht;
        
        // Create a new thread and flag the fact that the thread
        // did not get us any data yet. Then start the thread.
        Thread t = new Thread(this);
        _threadCompleted = false;
        _threadCanRunMethodExecute = true;
        t.start();

        // We calculate the current time that the thread started.
        // The delay to wait will be calculated below.  
        long base = System.currentTimeMillis();
        long delay = 0;
        synchronized(this) {
            try {
                while(t.isAlive() && _threadCompleted == false) {
                    delay = threadJavaCGIBridgeTimeOut - 
                        (System.currentTimeMillis() - base);
                    if (delay <= 0) break;
                    wait(delay);
                } // End of while
            } catch (InterruptedException e) { /* nothing */  }
        } // end of synchronized block
        t.stop(); // No effect if already done
                  // but stops if timed out...

        if (_threadCompleted == false) {
            throw new JavaCGIBridgeTimeOutException();
        }
        return _threadCGIData;
    }

    public void sendCGIDataOneWay(URL u) {
        sendCGIDataOneWay(u,null);
    } 

    public void sendCGIDataOneWay(URL u, Hashtable ht) {
        _threadURL = u;
        _threadFormVar = ht;
        Thread t= new Thread(this);
        _threadCompleted = false;
        _threadCanRunMethodExecute = true;
        t.start();
    }

    /**
     * Static method returns the default communication time out 
     * in milliseconds for the class.
     *
     * When the object retrieves data from a URL, it must get
     * the data within timeout milliseconds or a 
     * JavaCGIBridgeTimeOutException is thrown.
     *
     * @return default communication time out in milliseconds
     * @see #setDefaultThreadJavaCGIBridgeTimeOut
     */
    public static int getDefaultThreadJavaCGIBridgeTimeOut() { 
        return _defaultThreadJavaCGIBridgeTimeOut; 
    }

    /**
     * Static method sets the default communication time out in 
     * milliseconds for the class.
     *
     * When the object retrieves data from a URL, it must get
     * the data within timeout milliseconds or a 
     * JavaCGIBridgeTimeOutException is thrown.
     *
     * @param t default communication time out in milliseconds
     * @see #getDefaultThreadJavaCGIBridgeTimeOut
     */
    public static void setDefaultThreadJavaCGIBridgeTimeOut(int t) {
        _defaultThreadJavaCGIBridgeTimeOut = t;
    }

    /**
     * Returns the actual communication time out in milliseconds
     * for the object.
     *
     * When the object retrieves data from a URL, it must get
     * the data within timeout milliseconds or a
     * JavaCGIBridgeTimeOutException is thrown.
     * 
     * @return communication time out in milliseconds
     * @see #setThreadJavaCGIBridgeTimeOut
     * @see #getDefaultThreadJavaCGIBridgeTimeOut
     * @see #setDefaultThreadJavaCGIBridgeTimeOut
     */
    public int getThreadJavaCGIBridgeTimeOut() { 
        return threadJavaCGIBridgeTimeOut; 
    }

    /**
     * Sets the actual communication time out in milliseconds
     * for the object.
     *
     * When the object retrieves data from a URL, it must get
     * the data within timeout milliseconds or a
     * JavaCGIBridgeTimeOutException is thrown.
     *
     * @param t communication time out in milliseconds
     * @see #getThreadJavaCGIBridgeTimeOut
     * @see #getDefaultThreadJavaCGIBridgeTimeOut
     * @see #setDefaultThreadJavaCGIBridgeTimeOut
     */
    public void setThreadJavaCGIBridgeTimeOut(int t) { 
        threadJavaCGIBridgeTimeOut = t; 
    }

    /**
     * Sets the field separator for the object.  When getParsedData
     * method is called, the object uses the field separator to determine 
     * where fields in a returned record of the raw HTML result set
     * begin and end.
     *
     * @param s String containing new delimiting separator
     * @see #getParsedData
     * @see #getFieldSeparator
     */
    public void setFieldSeparator(String s) { _fieldSeparator = s; }

    /**
     * Returns the field separator for the object.  When getParsedData
     * method is called, the object uses the field separator to determine 
     * where fields in a returned record of the raw HTML result set
     * begin and end.
     *
     * @return Separator string
     * @see #getParsedData
     * @see #setFieldSeparator
     */
    public String getFieldSeparator() { return new String(_fieldSeparator); }

    /**
     * Sets the row separator for the object.  When getParsedData
     * method is called, the object uses the row separator to determine 
     * where rows/records of the raw HTML result set
     * begin and end.
     *
     * @param s String containing new delimiting separator
     * @see #getParsedData
     * @see #getRowSeparator
     */
    public void setRowSeparator(String s) { _recordSeparator = s; }

    /**
     * Returns the row separator for the object.  When getParsedData
     * method is called, the object uses the row separator to determine 
     * where rows/records of the raw HTML result set
     * begin and end.
     *
     * @return Separator string
     * @see #getParsedData
     * @see #setRowSeparator
     */
    public String getRowSeparator() { return new String(_recordSeparator); }

    /**
     * Sets the top separator for the object.  When getParsedData
     * method is called, the object uses the top separator to determine 
     * where the rows of data inside the raw HTML output actually begin.
     *
     * @param s String containing new delimiting separator
     * @see #getParsedData
     * @see #getTopSeparator
     */
    public void setTopSeparator(String s) { _startDataSeparator = s; }

    /**
     * Returns the top separator for the object.  When getParsedData
     * method is called, the object uses the top separator to determine
     * where the rows of data inside the raw HTML output actually begin.
     *
     * @return Separator string
     * @see #getParsedData
     * @see #setTopSeparator
     */
    public String getTopSeparator() { return new String(_startDataSeparator); }

    /**
     * Sets the bottom separator for the object.  When getParsedData
     * method is called, the object uses the bottom separator to determine
     * where the rows of data inside the raw HTML output actually end.
     *
     * @param s String containing new delimiting separator
     * @see #getParsedData
     * @see #getBottomSeparator
     */
    public void setBottomSeparator(String s) { _endDataSeparator = s; }

    /**
     * Returns the bottom separator for the object.  When getParsedData
     * method is called, the object uses the bottom separator to determine
     * where the rows of data inside the raw HTML output actually end.   
     *
     * @return Separator string
     * @see #getParsedData
     * @see #setBottomSeparator
     */
    public String getBottomSeparator() { return new String(_endDataSeparator); }

    /**
     * This run thread asynchronously POSTs and GETs data from
     * a URL and places the contents into the _threadCGIData variable.
     *
     * Since Threads do not return or pass parameter values, instance
     * variables in this object are used to maintain state.  This method
     * is only public so that the Thread class can launch the thread.
     *
     * @see #getRawCGIData
     */
    public void run() {
        if (_threadCanRunMethodExecute) {
            _threadCGIData = _getHttpRequestInThread(
                            _threadURL, _threadFormVar);
            _threadCompleted = true; // set completed flag
            synchronized(this) { notifyAll(); } 
              // ends the wait() for thread to end in getRawCGIData
        }
    } // end of run -- The thread runs

    /** 
     * Returns the form variables inside the hashtable as a 
     * URL encoded string of parameters for a CGI based program
     * to process.
     *
     * @param ht Hashtable containing form variables
     * @return URLencoded string of HTML form variables & values
     * @see #addFormValue
     */
    private String _getURLEncodedHashTable(Hashtable ht) {
        StringBuffer encodedString = null;
        Vector vFormValues = null;

        // First, we enumerate through the keys (Form variables)
        for (Enumeration eKeys = ht.keys() ; eKeys.hasMoreElements() ;) {
            String formVariable = (String)eKeys.nextElement();

            vFormValues = (Vector)ht.get(formVariable);
            // Now, we have to enumerate through the values for the form variables.
            //
            // NOTE: It IS entirely possible that a form variable has MANY values
            // to simulate events such as multiple checkboxes or multiple select <SELECT>
            // form elements in an HTML form.
            for (Enumeration eVals = vFormValues.elements() ; eVals.hasMoreElements() ;) {
                String formValue = (String)eVals.nextElement();

                if (encodedString != null) {
                    encodedString.append("&");
                } else {
                    encodedString = new StringBuffer();
                }

                encodedString.append(URLEncoder.encode(formVariable) + "=" +
                    URLEncoder.encode(formValue));
            }
        }
        return encodedString.toString();
    } // End of _getURLEncodedHashTable

    /**
     * Returns the HTTP Request data to the thread that
     * was launched to GET/POST data for a URL.  This is
     * a private method.
     *
     * @param u URL to retrieve and post data for
     * @param ht Form variables to send to URL
     * @return String containing retrieved HTML text
     * @see #run
     */
    private String _getHttpRequestInThread(URL u, Hashtable ht) {
        String postContent = null;
        StringBuffer strBuffer = new StringBuffer();


        if (ht != null) {
            postContent = _getURLEncodedHashTable(ht);
        }

        try {
            URLConnection       urlConn;
            DataOutputStream    dos;
            DataInputStream     dis;


            // Establish the URL connection
            urlConn = u.openConnection();

            // We always want some input from the CGI script
            // in general even if it is just a success story
            urlConn.setDoInput (true);

            // If ht was not null, then we know there is
            // data to be posted.
            if (ht != null) {
                urlConn.setDoOutput(true);
            }

            // Turn off caching as this may screw with
            // our dynamic programs

            urlConn.setUseCaches (false);
            // Specify the content type
            urlConn.setRequestProperty
               ("Content-type", "application/x-www-form-urlencoded");

            // Send the POST data if we are writing
            if (ht != null) {
                dos = new DataOutputStream(urlConn.getOutputStream());
                dos.writeBytes(postContent);
                dos.flush();
                dos.close();
            }

            dis = new DataInputStream (urlConn.getInputStream());

            String str;

            if (_parseDataAsVector == true) {
                _vectorParsedData = new Vector();
            }
            int startDataIndex = -1, endDataIndex = -1; 
            int lastRecordIndex = 0;
            int lastFoundIndex = 0;
            boolean stopParsingRecords = false;

            while (null != ((str = dis.readLine())))
            {
                // Do troubleshooting here such as System.println

                strBuffer.append(str + "\n");

                int vectorLineCount = 0;
                int htmlLineCount = 0;

                if (_parseDataAsVector == true) {
                    if (startDataIndex == -1) {
                        startDataIndex = _getIndex(strBuffer,_startDataSeparator,0,-1);
                        if (startDataIndex != -1) 
                            lastFoundIndex = startDataIndex + _startDataSeparator.length();
                    }

                    if (endDataIndex == -1 && startDataIndex != -1) 
                        endDataIndex = _getIndex(strBuffer,_endDataSeparator,lastFoundIndex,-1);

                    if (endDataIndex != -1 && 
                        lastRecordIndex + _recordSeparator.length() >= endDataIndex) 
                          stopParsingRecords = true;

                    if (startDataIndex != -1 && !stopParsingRecords) { // We can begin parsing!
                        int recordIndex = _getIndex(strBuffer,_recordSeparator,lastFoundIndex,-1);
                        int startIndex = lastFoundIndex;
                        // Take care of case where _recordSeparator was forgotten
                        // before _endDataSeparator.
                        if ((recordIndex == -1 && endDataIndex != -1 && 
                            (endDataIndex - lastRecordIndex) != _recordSeparator.length()) ||
                            (recordIndex != -1 && recordIndex > endDataIndex && endDataIndex != -1)) {
                            recordIndex = endDataIndex;
                            lastFoundIndex = recordIndex + _endDataSeparator.length();
                            stopParsingRecords = true;
                        } else if (recordIndex != -1) {
                            lastFoundIndex = recordIndex + _recordSeparator.length();
                        }
                        if (recordIndex != -1) {
                            lastRecordIndex = recordIndex;
                            _vectorParsedData.addElement(new Vector());
                            int fieldIndex = startIndex;
                            while (fieldIndex != -1) {
                                int beginFieldIndex = fieldIndex;
                                fieldIndex = _getIndex(strBuffer,_fieldSeparator,fieldIndex,recordIndex);
                                if (fieldIndex != -1) {
                                    char c [] = new char[fieldIndex - beginFieldIndex];
                                    strBuffer.getChars(beginFieldIndex,fieldIndex,c,0);
                                    fieldIndex = fieldIndex + _fieldSeparator.length();
                                     ((Vector)_vectorParsedData.
                                       lastElement()).addElement(new String(c)); 
                                } else {
                                    char c [] = new char[recordIndex - beginFieldIndex];
                                    strBuffer.getChars(beginFieldIndex,recordIndex,c,0);
                                     ((Vector)_vectorParsedData.
                                       lastElement()).addElement(new String(c)); 
                                }
                            } // Processing fields in a record
                            vectorLineCount++;
                            if (_numberOfRecordsToNotifyObservers > 0 && 
                                vectorLineCount % _numberOfRecordsToNotifyObservers == 0)
                                notifyObservers(_vectorParsedData);
                        } // Processing recordIndex

                    }
                }

                htmlLineCount++;
                if (_numberOfHtmlLinesToNotifyObserver > 0 &&
                    htmlLineCount % _numberOfHtmlLinesToNotifyObserver == 0)
                    notifyObservers(strBuffer);

            } // end of while reading lines from HTTP connection

            dis.close ();

            if (_numberOfHtmlLinesToNotifyObserver > 0 ||
                _notifyObserversOfEndOfHtml)
                notifyObservers(strBuffer);

            if (_numberOfRecordsToNotifyObservers > 0 ||
                _notifyObserversOfEndOfVector)
                notifyObservers(_vectorParsedData);

            if (_notifyObserversOfEndOfData) {
                notifyObservers();
            }
            
        }   catch (MalformedURLException me)
            {
                System.err.println("MalformedURLException: " + me);
            }
            catch (IOException ioe)
            {
                System.err.println("IOException: " + ioe.getMessage());
            }

            return strBuffer.toString();

    } // end of _getHttpRequestInThread

    private int _getIndex(StringBuffer sb, String toLookFor, int begin, int end) {
    // Code here is based loosely on String.indexOf code in java.lang.String source

        if (end == -1) end = sb.length();
        if (begin >= end) return -1;

        char v1[] = new char[end - begin];
        sb.getChars(begin,end,v1,0);
        char v2[] = toLookFor.toCharArray();

        int max = v1.length - v2.length;

      test:
        for (int i = 0; i <= max ; i++) {
            int n = v2.length;
            int j = i;
            int k = 0;
            while (n-- != 0) {
                if (v1[j++] != v2[k++]) {
                    continue test;
                }
            }
            // Must return the correct offset...
            return begin + i;
       }
       return -1;
    } // end of _getIndex


} // End of JavaCGIBridge class
